/*
   orxonox - the future of 3D-vertical-scrollers

   Copyright (C) 2004 orx

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   ### File Specific:
   main-programmer: bottac@ee.ethz.ch
*/

#include "height_map.h"
#include "model.h"
#include "texture.h"
#include "vector.h"
#include "material.h"
#include "p_node.h"
#include "state.h"
#include "resource_manager.h"
#include "debug.h"

// INCLUDING SDL_Image
#ifdef HAVE_SDL_IMAGE_H
#include <SDL_image.h>
#else
#include <SDL/SDL_image.h>
#endif

HeightMap::HeightMap()
{

}

HeightMap::HeightMap(const char* height_map_name = NULL) : VertexArrayModel()
{
   this->setClassID(CL_HEIGHT_MAP, "HeightMap");
   heightMap =  IMG_Load(height_map_name);
   if(heightMap!=NULL) {

		 PRINTF(0)("loading Image %s\n", height_map_name);
		 PRINTF(0)("width : %i\n", heightMap->w);
		 PRINTF(0)("height : %i\n", heightMap->h);
		 PRINTF(0)("%i Byte(s) per Pixel \n", heightMap->format->BytesPerPixel);
		 PRINTF(0)("Rshift : %i\n", heightMap->format->Rshift);
		 PRINTF(0)("Bshift: %i\n", heightMap->format->Bshift);
		 PRINTF(0)("Gshift: %i\n", heightMap->format->Gshift);
		 PRINTF(0)("Rmask: %i\n", heightMap->format->Rmask);
		 PRINTF(0)("Gmask: %i\n", heightMap->format->Gmask);
		}

     else       PRINTF(4)("oops! couldn't load %s for some reason.\n", height_map_name);

	
  generateNormalVectorField();

  shiftX = 0;
  shiftY = 0;
  shiftZ = 0;

}

HeightMap::HeightMap(const char* height_map_name = NULL, const char* colour_map_name = NULL) : VertexArrayModel()
{
this->setClassID(CL_HEIGHT_MAP, "HeightMap");

   heightMap =  IMG_Load(height_map_name);
   if(heightMap!=NULL) {

		 PRINTF(0)("loading Image %s\n", height_map_name);
		 PRINTF(0)("width : %i\n", heightMap->w);
		 PRINTF(0)("height : %i\n", heightMap->h);
		 PRINTF(0)("%i Byte(s) per Pixel \n", heightMap->format->BytesPerPixel);
		 PRINTF(0)("Rshift : %i\n", heightMap->format->Rshift);
		 PRINTF(0)("Bshift: %i\n", heightMap->format->Bshift);
		 PRINTF(0)("Gshift: %i\n", heightMap->format->Gshift);
		 PRINTF(0)("Rmask: %i\n", heightMap->format->Rmask);
		 PRINTF(0)("Gmask: %i\n", heightMap->format->Gmask);
		}

     else       PRINTF(4)("oops! couldn't load %s for some reason.\n", height_map_name);


    generateNormalVectorField();


  if(colour_map_name != NULL)
  {
  colourMap = IMG_Load(colour_map_name);
  if(colourMap != NULL)
  		{
  		 PRINTF(0)("loading Image %s\n", colour_map_name);
		 PRINTF(0)("width : %i\n", colourMap->w);
		 PRINTF(0)("height : %i\n", colourMap->h);
		 PRINTF(0)("%i Byte(s) per Pixel \n", colourMap->format->BytesPerPixel);
		 PRINTF(0)("Rshift : %i\n", colourMap->format->Rshift);
		 PRINTF(0)("Bshift: %i\n", colourMap->format->Bshift);
		 PRINTF(0)("Gshift: %i\n", colourMap->format->Gshift);
		 PRINTF(0)("Rmask: %i\n", colourMap->format->Rmask);
		 PRINTF(0)("Gmask: %i\n", colourMap->format->Gmask);
		 }
		 else       PRINTF(4)("oops! couldn't load %s for some reason.\n", colour_map_name);

		 
		   
  if(colourMap != NULL)
  {
  colours = (unsigned char *) colourMap->pixels;
  hasColourMap = true;
  }
  else hasColourMap = false;
  
	
  shiftX = 0;
  shiftY = 0;
  shiftZ = 0;
  }


}
HeightMap::~HeightMap()
{
	delete heightMap;
}

void HeightMap::load()
{

unsigned char height = 0;
 int offset = 0;
 int g = 0;
 

 
 heights  = (unsigned char*) heightMap->pixels;



 
 if(heightMap != NULL && heightMap->format->BitsPerPixel == 8 )
	
	SDL_LockSurface(heightMap);

     
	for(int i = 0 ; i < heightMap->h -sampleRate ; i +=sampleRate)
	{
		int j = 0;

		float r;
		float g;
		float b;
	    for(int j = 0 ; j < heightMap->w -sampleRate  ;  j += sampleRate)
	    {
	    	 // To be fixed
		 r = (float)colours[3*j+2 + 3*i*(heightMap->w )];
		 g = (float)colours[3*j+1 + 3*i*(heightMap->w)];
		 b = (float)colours[3*j+0 + 3*i*(heightMap->w)];
		
			height =  heights[j+sampleRate + i*(heightMap->w )];
		this->addVertex(scaleX*(heightMap->h -i) + shiftX ,((double)(height)*scaleY) + shiftY ,scaleZ*(j+sampleRate) + shiftZ); // Top Right

	
		this->addNormal(normalVectorField[i][j+sampleRate].y,normalVectorField[i][j+sampleRate].z,normalVectorField[i][j+sampleRate].x);
		this->addTexCoor(((j/sampleRate)%texRate)/texRatef,((i/sampleRate)%texRate)/texRatef);

 		this->addColor(r/255.0,g/255.0,b/255.0);


	    }
		
	    
		 j = (heightMap->w/sampleRate - sampleRate);

		this->addVertex(scaleX*(heightMap->h -i)+ shiftX,-1000.0,scaleZ*(j+sampleRate)+ shiftZ); // Top Right

		this->addNormal(normalVectorField[i][j+sampleRate].y,normalVectorField[i][j+sampleRate].z,normalVectorField[i][j+sampleRate].x);
		this->addTexCoor(((j/sampleRate)%4)/4.0,((i/sampleRate)%4)/4.0);
 		this->addColor(0.1,0.1,0.9);
					

		

		}

	SDL_UnlockSurface(heightMap);


	int cnt = 0;
	for(int i = 0 ; i < heightMap->h - 2*sampleRate ; i +=sampleRate)
	{
		
	    for(int j = 0 ; j < heightMap->w - sampleRate  ;  j += sampleRate)
	    {
		
	    	this->addIndice(cnt);
		this->addIndice(cnt  + (heightMap->w)/ sampleRate );
		cnt++;
		
	    }
	    	
	    	cnt++;
		 this->newStripe();	
		
		
	}

	this->finalize();


}




void HeightMap::generateNormalVectorField()
{
int delta = 1;
heights  = (unsigned char*) heightMap->pixels;

//Create a Dynamicly sized 2D-Array to store our normals
normalVectorField =  new Vector* [heightMap->h];
for(int i=0;i<heightMap->h;i++)
normalVectorField [i]= new (Vector [heightMap->w]);



// !!! Does not yet calculate any normals for some border points!!!!!

if(heightMap != NULL && heightMap->format->BitsPerPixel == 8 )
	{
	SDL_LockSurface(heightMap);
	for(int i = 0 ; i < heightMap->h - 1  ; i ++)
	{
	    for(int j = 0; j < heightMap->w  - 1  ;  j ++)
	    {


		delta = (int)heights[j + (i+1)*(heightMap->w )] -  (int) heights[j + i*(heightMap->w )];
		Vector a =  Vector(-scaleX,(float)delta*scaleY  ,0.0f);

		delta = (int)heights[j+1 + i*(heightMap->w )] - (int)heights[j + i*(heightMap->w )];
		Vector b =  Vector(0.0f,(float) delta*scaleY ,scaleZ);


		 normalVectorField[i][j] = b.cross(a);
		 normalVectorField[i][j].normalize();
		  
	     }
	}
	SDL_UnlockSurface(heightMap);

}

}





void HeightMap::scale(Vector v)
{
 scaleX = v.x;
 scaleY = v.y;
 scaleZ = v.z;
}

void HeightMap::shift(Vector v)
{
 shiftX = v.x;
 shiftY = v.y;
 shiftZ = v.z;
}


float HeightMap::getHeight(float x, float y)
{
 int xInt = (int)x / scaleX;  x -= (float)xInt*scaleX; xInt = heightMap->h - xInt;
 int yInt = (int)y / scaleZ;    y -= (float)yInt*scaleZ;
 if(xInt <= 0 || xInt >= heightMap->h || yInt <= 0 || yInt >= heightMap->w  ) return 0.0f;
 float height = heights[yInt + (xInt)*heightMap->w]*scaleY;
 float a = normalVectorField[(xInt)][yInt].x;
 float b = normalVectorField [(xInt)][yInt].z;
 float c = normalVectorField [(xInt)][yInt].y;
  height -= ( (a/c)*(x) + (b/c)*(y))*scaleY;
 return height;
}
